#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
Initialize the library.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from . import attribute
from . import filters
from . import rockphysics
from . import reflection
from . import noise
from . import transform
from . import unit
from . import util

__all__ = ["attribute",
           "filters",
           "rockphysics",
           "reflection",
           "noise",
           "transform",
           "unit",
           "util"]

__version__ = "unknown"
try:
    from _version import __version__
except ImportError:
    pass



from numpy import zeros
import numpy as np
from bruges.attribute import spectra


def spectraldecomp(data, f=(.1,.25,.4),
                   window_length=32, dt=1,
                   window_type='hann',
                   overlap=1,
                   normalize=False):
    """
    Uses the STFT to decompose traces into normalized spectra. Only
    frequencies defined by the user will be output. Using 3
    frequencies will work for RGB color plots.

    :param data: A 1/2D array (samples, traces) of data that will
                 be decomposed.
    :keyword f: A list of frequencies to select from the spectra.
    :keyword window_length: The length of fft window to use for
                            the STFTs. Defaults to 32. Can be
                            provided in seconds if dt is specified.
    :keyword dt: The time sample interval of the traces.
    :keyword window_type: The type of window to use for the STFT. The
                          same input as scipy.signal.get_window.
    :keyword overlap: The fraction of overlap between adjacent
                      STFT windows

    :keyword normalize: Normalize the energy in each STFT window

    :returns: an array of shape (samples, traces, f)
    """

    
    # Do the 1D case
    if len(data.shape) == 1:
        ntraces = 1
    else:
        ntraces = data.shape[-1]

    if overlap > 1: overlap = 1
        

    zp = 4*window_length
    # TODO We should think about removing these for loops
    for i in range(ntraces):

        if(ntraces == 1):
            spect = spectra(data, window_length, dt=dt,
                            window_type=window_type, overlap=overlap,
                            normalize=normalize, zero_padding=zp)
            if( i == 0 ): output = np.zeros( (spect.shape[0], len(f)))
            
        else:
            spect = spectra(data[:,i], window_length, dt=dt,
                            window_type=window_type, overlap=overlap,
                            normalize=normalize, zero_padding=zp)
            if( i == 0 ):
                output = np.zeros( (spect.shape[0],ntraces,
                                   len(f) ))

        res = ((1. / dt) /2.) / spect.shape[-1]

        # TODO again, we should think about removing this loop
        for j in range(len(f)):

            index = int(f[j] / res)

            if( ntraces == 1 ): output[:,j] = spect[:,index] 
            else: output[:,i, j] = spect[:,index]


        
    return( output )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Various reflectivity algorithms.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from numpy import tan, sin, cos

from ..rockphysics import pr


def scattering_matrix(vp1, vs1, rho1, vp0, vs0, rho0, theta1):
    '''
    Full Zoeppritz solution, considered the definitive solution.
    Calculates the angle dependent p-wave reflectivity of an interface
    between two mediums.

    Written by: Wes Hamlyn

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp0: The p-wave velocity of the lower medium.
    :param vs0: The s-wave velocity of the lower medium.
    :param rho0: The density of the lower medium.

    :param theta1: A scalar  [degrees].

    :returns: a 4x4 array representing the scattering matrix
                at the incident angle theta1.
    '''
    # Make sure theta1 is an array
    theta1 = np.radians(np.array(theta1))
    if theta1.size == 1:
        theta1 = np.expand_dims(theta1, axis=1)

    # Set the ray paramter, p
    p = sin(theta1) / vp1  # ray parameter

    # Calculate reflection & transmission angles for Zoeppritz
    theta2 = np.arcsin(p * vp0)      # Trans. angle of P-wave
    phi1 = np.arcsin(p * vs1)     # Refl. angle of converted S-wave
    phi2 = np.arcsin(p * vs0)      # Trans. angle of converted S-wave

    # Matrix form of Zoeppritz Equations... M & N are matricies
    M = np.array([[-sin(theta1), -cos(phi1), sin(theta2), cos(phi2)],
                  [cos(theta1), -sin(phi1), cos(theta2), -sin(phi2)],
                  [2 * rho1 * vs1 * sin(phi1) * cos(theta1),
                   rho1 * vs1 * (1 - 2 * sin(phi1) ** 2),
                   2 * rho0 * vs0 * sin(phi2) * cos(theta2),
                   rho0 * vs0 * (1 - 2 * sin(phi2) ** 2)],
                  [-rho1 * vp1 * (1 - 2 * sin(phi1) ** 2),
                   rho1 * vs1 * sin(2 * phi1),
                   rho0 * vp0 * (1 - 2 * sin(phi2) ** 2),
                   -rho0 * vs0 * sin(2 * phi2)]], dtype='float')

    N = np.array([[sin(theta1), cos(phi1), -sin(theta2), -cos(phi2)],
                  [cos(theta1), -sin(phi1), cos(theta2), -sin(phi2)],
                  [2 * rho1 * vs1 * sin(phi1) * cos(theta1),
                   rho1 * vs1 * (1 - 2 * sin(phi1) ** 2),
                   2 * rho0 * vs0 * sin(phi2) * cos(theta2),
                   rho0 * vs0 * (1 - 2 * sin(phi2) ** 2)],
                  [rho1 * vp1 * (1 - 2 * sin(phi1) ** 2),
                   -rho1 * vs1 * sin(2 * phi1),
                   - rho0 * vp0 * (1 - 2 * sin(phi2) ** 2),
                   rho0 * vs0 * sin(2 * phi2)]], dtype='float')

    zoep = np.zeros((4, 4, M.shape[-1]))
    for i in range(M.shape[-1]):
        Mi = M[..., i]
        Ni = N[..., i]
        dt = np.dot(np.linalg.inv(Mi), Ni)
        zoep[..., i] = dt

    return zoep


def zoeppritz_element(vp1, vs1, rho1, vp0, vs0, rho0, theta1, element='PdPu'):
    """
    Returns any mode reflection coefficients from the Zoeppritz
    scattering matrix. Pass in the mode as element, e.g. 'PdSu' for PS.

    Wraps scattering_matrix().

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    elements = np.array([['PdPu', 'SdPu', 'PuPu', 'SuPu'],
                         ['PdSu', 'SdSu', 'PuSu', 'SuSu'],
                         ['PdPd', 'SdPd', 'PuPd', 'SuPd'],
                         ['PdSd', 'SdSd', 'PuSd', 'SuSd']])

    Z = scattering_matrix(vp1, vs1, rho1, vp0, vs0, rho0, theta1)

    return np.squeeze(Z[np.where(elements == element)])


def zoeppritz(vp1, vs1, rho1, vp0, vs0, rho0, theta1):
    '''
    Returns the PP reflection coefficients from the Zoeppritz
    scattering matrix.

    Wraps zoeppritz_element().

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    '''
    return zoeppritz_element(vp1, vs1, rho1, vp0, vs0, rho0, theta1, 'PdPu')


def zoeppritz_rpp(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Exact Zoeppritz from expression.

    This is useful because we can pass arrays to it, which we can't do to
    scattering_matrix().

    Dvorkin et al. (2014). Seismic Reflections of Rock Properties. Cambridge.
    """
    theta1 = np.radians(theta1)
    p = np.sin(theta1) / vp1  # Ray parameter
    theta2 = np.arcsin(p * vp2)
    phi1 = np.arcsin(p * vs1)  # Reflected S
    phi2 = np.arcsin(p * vs2)  # Transmitted S

    a = rho2 * (1 - 2 * np.sin(phi2)**2.) - rho1 * (1 - 2 * np.sin(phi1)**2.)
    b = rho2 * (1 - 2 * np.sin(phi2)**2.) + 2 * rho1 * np.sin(phi1)**2.
    c = rho1 * (1 - 2 * np.sin(phi1)**2.) + 2 * rho2 * np.sin(phi2)**2.
    d = 2 * (rho2 * vs2**2 - rho1 * vs1**2)

    E = (b * np.cos(theta1) / vp1) + (c * np.cos(theta2) / vp2)
    F = (b * np.cos(phi1) / vs1) + (c * np.cos(phi2) / vs2)
    G = a - d * np.cos(theta1)/vp1 * np.cos(phi2)/vs2
    H = a - d * np.cos(theta2)/vp2 * np.cos(phi1)/vs1

    D = E*F + G*H*p**2

    rpp = (1/D) * (F*(b*(np.cos(theta1)/vp1) - c*(np.cos(theta2)/vp2))\
          - H*p**2 * (a + d*(np.cos(theta1)/vp1)*(np.cos(phi2)/vs2)))

    return rpp


def akirichards(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    This is the formulation from Avseth et al.,
    Quantitative seismic interpretation,
    Cambridge University Press, 2006. Adapted for a 4-term formula.
    See http://subsurfwiki.org/wiki/Aki-Richards_equation

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle
    """

    # We are not using this for anything, but will
    # critical_angle = arcsin(vp1/vp2)

    # Do we need to ensure that we get floats out before
    # computing sines?
    if np.ndim(vp1) == 0:
        vp1 = float(vp1)
    else:
        vp1 = np.array(vp1).astype(float)

    theta1 = np.radians(theta1)
    theta2 = np.arcsin(vp2/vp1*sin(theta1))

    # Compute the various parameters
    drho = rho2-rho1
    dvp = vp2-vp1
    dvs = vs2-vs1
    meantheta = (theta1+theta2) / 2.0
    rho = (rho1+rho2) / 2.0
    vp = (vp1+vp2) / 2.0
    vs = (vs1+vs2) / 2.0

    # Compute the coefficients
    w = 0.5 * drho/rho
    x = 2 * (vs/vp1)**2 * drho/rho
    y = 0.5 * (dvp/vp)
    z = 4 * (vs/vp1)**2 * (dvs/vs)

    # Compute the terms
    term1 = w
    term2 = -1 * x * sin(theta1)**2
    term3 = y / cos(meantheta)**2
    term4 = -1 * z * sin(theta1)**2

    if terms:
        return term1, term2, term3, term4
    else:
        return (term1 + term2 + term3 + term4)


def akirichards_alt(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    This is another formulation of the Aki-Richards solution.
    See http://subsurfwiki.org/wiki/Aki-Richards_equation

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """

    # We are not using this for anything, but will
    # critical_angle = arcsin(vp1/vp2)

    # Do we need to ensure that we get floats out before
    # computing sines?
    if np.ndim(vp1) == 0:
        vp1 = float(vp1)
    else:
        vp1 = np.array(vp1).astype(float)

    theta1 = np.radians(theta1)
    theta2 = np.arcsin(vp2/vp1*sin(theta1))

    # Compute the various parameters
    drho = rho2-rho1
    dvp = vp2-vp1
    dvs = vs2-vs1
    theta = (theta1+theta2)/2.0
    rho = (rho1+rho2)/2.0
    vp = (vp1+vp2)/2.0
    vs = (vs1+vs2)/2.0

    # Compute the three terms
    term1 = 0.5 * (dvp/vp + drho/rho)
    term2 = (0.5*dvp/vp-2*(vs/vp)**2*(drho/rho+2*dvs/vs)) * sin(theta)**2
    term3 = 0.5 * dvp/vp * (tan(theta)**2 - sin(theta)**2)

    if terms:
        return term1, term2, term3
    else:
        return (term1 + term2 + term3)

def fatti(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Compute reflectivities with Fatti's formulation of the
    Aki-Richards equation, which does not account for the
    critical angle. Fatti et al (1994), Geophysics 59 (9).

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    # Do we need to ensure that we get floats out before computing
    # sines?
    if np.ndim(vp1) == 0:
        vp1 = float(vp1)
    else:
        vp1 = np.array(vp1).astype(float)

    theta1 = np.radians(theta1)

    # Compute the various parameters
    drho = rho2-rho1
    rho = (rho1+rho2) / 2.0
    vp = (vp1+vp2) / 2.0
    vs = (vs1+vs2) / 2.0
    dip = (vp2*rho2 - vp1*rho1)/(vp2*rho2 + vp1*rho1)
    dis = (vs2*rho2 - vs1*rho1)/(vs2*rho2 + vs1*rho1)
    d = drho/rho

    # Compute the three terms
    term1 = (1 + tan(theta1)**2) * dip
    term2 = -8 * (vs/vp)**2 * sin(theta1)**2 * dis
    term3 = -1 * (0.5 * tan(theta1)**2 - 2 * (vs/vp)**2 * sin(theta1)**2) * d

    if terms:
        return term1, term2, term3
    else:
        return (term1 + term2 + term3)

def shuey(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Compute Shuey approximation with 3 terms.
    http://subsurfwiki.org/wiki/Shuey_equation

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    theta1 = np.radians(theta1)

    # Compute some parameters
    drho = rho2-rho1
    dvp = vp2-vp1
    dvs = vs2-vs1
    rho = (rho1+rho2)/2.0
    vp = (vp1+vp2)/2.0
    vs = (vs1+vs2)/2.0

    # Compute three-term reflectivity

    r0 = 0.5 * (dvp/vp + drho/rho)
    g = 0.5 * dvp/vp - 2 * (vs**2/vp**2) * (drho/rho + 2 * dvs/vs)
    f = 0.5 * dvp/vp

    term1 = r0
    term2 = g * np.sin(theta1)**2
    term3 = f * (np.tan(theta1)**2 - np.sin(theta1)**2)

    if terms:
        return term1, term2, term3
    else:
        return (term1 + term2 + term3)

def shuey2(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Compute Shuey approximation with 2 terms.

    Wraps shuey().
    """
    r0, rg, rf = shuey(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=True)

    if terms:
        return r0, rg
    else:
        return r0 + rg

def shuey3(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Compute Shuey approximation with 2 terms.

    Wraps shuey().
    """
    r0, rg, rf = shuey(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=True)
    
    if terms:
        return r0, rg, rf
    else:
        return r0 + rg + rf

def bortfeld2(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    The 2-term Bortfeld approximation for ava analysis.

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    theta1 = np.radians(theta1)
    # Bortfeld only needs one extra parameter
    theta2 = np.arcsin(vp2/vp1*np.sin(theta1))  # radians

    # This breaks if theta = 90 deg
    term1 = 0.5 * np.log((vp2*rho2*cos(theta1)) / (vp1*rho1*cos(theta2)))

    svp2 = (np.sin(theta1)/vp1)**2
    dvs2 = (vs1**2-vs2**2)
    term2 = svp2 * dvs2 * (2+np.log(rho2/rho1)/np.log(vs2/vs1))

    if terms:
        return term1, term2
    else:
        return (term1 + term2)

def bortfeld3(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Compute Bortfeld approximation with three terms.
    http://sepwww.stanford.edu/public/docs/sep111/marie2/paper_html/node2.html

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    theta1 = np.radians(theta1)

    # Compute some parameters
    drho = rho2-rho1
    dvp = vp2-vp1
    dvs = vs2-vs1
    rho = (rho1+rho2)/2.0
    vp = (vp1+vp2)/2.0
    vs = (vs1+vs2)/2.0
    k = (2 * vs/vp)**2
    rsh = 0.5 * (dvp/vp - k*drho/rho - 2*k*dvs/vs)

    # Compute three-term reflectivity
    term1 = 0.5 * (dvp/vp + drho/rho)
    term2 = rsh * np.sin(theta1)**2
    term3 = 0.5 * dvp/vp * np.tan(theta1)**2 * np.sin(theta1)**2

    if terms:
        return term1, term2, term3
    else:
        return (term1 + term2 + term3)

def hilterman(vp1, vs1, rho1, vp2, vs2, rho2, theta1, terms=False):
    """
    Hilterman (1989) approximation.
    According to Dvorkin: "arguably the simplest and a very convenient
    [approximation]." At least for small angles and small contrasts.

    :param vp1: The p-wave velocity of the upper medium.
    :param vs1: The s-wave velocity of the upper medium.
    :param rho1: The density of the upper medium.

    :param vp2: The p-wave velocity of the lower medium.
    :param vs2: The s-wave velocity of the lower medium.
    :param rho2: The density of the lower medium.

    :param theta1: An array of incident angles to use for reflectivity
                   calculation [degrees].

    :returns: a vector of len(theta1) containing the reflectivity
             value corresponding to each angle.
    """
    theta1 = np.radians(theta1)
    ip1 = vp1 * rho1
    ip2 = vp2 * rho2
    rpp0 = (ip2 - ip1) / (ip2 + ip1)
    dpr = pr(vp2, vs2) - pr(vp1, vs1)

    term1 = rpp0
    term2 = 2.25*(dpr - rpp0)*np.sin(theta1)**2

    if terms:
        return term1, term2
    else:
        return (term1 + term2)



from .wavelets import ricker
from .wavelets import ormsby
from .wavelets import sweep
from .wavelets import rotate_phase
from .kernels import gaussian_kernel
from .filters import snn

__all__ = ["ricker",
           "ormsby",
	   "sweep",
	   "rotate_phase",
	   "gaussian_kernel",
	   "snn"
	   ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from .energy import energy
from .similarity import similarity
from .dipsteer import dipsteer
from .spectrogram import spectra
from .spectraldecomp import spectraldecomp

__all__ = ["energy",
           "similarity",
	   "dipsteer",
	   "spectra",
	   "spectraldecomp",
	   ]



# -*- coding: utf-8 -*-
'''
===================
moduli.py
===================

Converts between various acoustic/eslatic parameters,
and provides a way to calculate all the elastic moduli
from Vp, Vs, and rho

Created June 2014

@author: Matt Hall

Using equations http://www.subsurfwiki.org/wiki/Elastic_modulus
from Mavko, G, T Mukerji and J Dvorkin (2003), The Rock Physics Handbook, Cambridge University Press

'''
import numpy as np


def youngs(vp=None, vs=None, rho=None, mu=None, lam=None, bulk=None, pr=None, pmod=None):
    '''
    Computes Young's modulus given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. lambda and mu, or bulk and P
    moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from lam, mu, bulk, pr, and pmod

    Returns:
        Young's modulus in pascals, Pa

    '''
    if vp and vs and rho:
        return rho * vs**2 * (3.*vp**2 - 4.*vs**2) / (vp**2 - vs**2)

    elif mu and lam:
        return mu * (3.*lam + 2*mu) / (lam + mu)

    elif bulk and lam:
        return 9.*bulk * (bulk - lam) / (3.*bulk - lam)

    elif bulk and mu:
        return 9.*bulk*mu / (3.*bulk + mu)

    elif lam and pr:
        return lam * (1+pr) * (1 - 2*pr) / pr

    elif pr and mu:
        return 2. * mu * (1+pr)

    elif pr and bulk:
        return 3. * bulk * (1 - 2*pr)

    else:
        return None


def bulk(vp=None, vs=None, rho=None, mu=None, lam=None, youngs=None, pr=None, pmod=None):
    '''
    Computes bulk modulus given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. lambda and mu, or Young's
    and P moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from lam, mu, youngs, pr, and pmod

    Returns:
        Bulk modulus in pascals, Pa

    '''

    if vp and vs and rho:
        return rho * (vp**2 - (4./3.)*(vs**2))

    elif mu and lam:
        return lam + 2*mu/3.

    elif mu and youngs:
        return youngs * mu / (9.*mu - 3.*youngs)

    elif lam and pr:
        return lam * (1+pr) / 3.*pr

    elif pr and mu:
        return 2. * mu * (1+pr) / (3. - 6.*pr)

    elif pr and youngs:
        return youngs / (3. - 6.*pr)

    else:
        return None


def pr(vp=None, vs=None, rho=None, mu=None, lam=None, youngs=None, bulk=None, pmod=None):
    '''
    Computes Poisson ratio given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. lambda and mu, or Young's
    and P moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from lam, mu, youngs, bulk, and pmod

    Returns:
        Poisson's ratio, dimensionless

    '''


    if vp and vs:
        return (vp**2. - 2.*vs**2) / (2. * (vp**2 - vs**2))

    elif mu and lam:
        return lam / (2. * (lam+mu))

    elif mu and youngs:
        return (youngs / (2.*mu)) - 1

    elif lam and bulk:
        return lam / (3.*bulk - lam)

    elif bulk and mu:
        return (3.*bulk - 2*mu) / (6.*bulk + 2*mu)

    elif bulk and youngs:
        return (3.*bulk - youngs) / (6.*bulk)

    else:
        return None


def mu(vp=None, vs=None, rho=None, pr=None, lam=None, youngs=None, bulk=None, pmod=None):
    '''
    Computes shear modulus given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. lambda and bulk, or Young's
    and P moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from lam, bulk, youngs, pr, and pmod

    Returns:
        Shear modulus in pascals, Pa

    '''


    if vs and rho:
        return rho * vs**2

    elif bulk and lam:
        return 3. * (bulk - lam) / 2.

    elif bulk and youngs:
        return 3. * bulk * youngs / (9.*bulk - youngs)

    elif lam and pr:
        return lam * (1 - 2.*pr) / (2.*pr)

    elif pr and youngs:
        return youngs / (2. * (1 + pr))

    elif pr and bulk:
        return 3. * bulk * (1 - 2*pr) / (2. * (1 + pr))

    else:
        return None


def lam(vp=None, vs=None, rho=None, pr=None,  mu=None, youngs=None, bulk=None, pmod=None):
    '''
    Computes lambda given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. bulk and mu, or Young's
    and P moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from bulk, mu, youngs, pr, and pmod

    Returns:
        Lambda in pascals, Pa

    '''


    if vs and rho:
        return rho * (vp**2 - 2.*vs**2.)

    elif youngs and mu:
        return mu * (youngs - 2.*mu) / (3.*mu - youngs)

    elif bulk and mu:
        return bulk - (2.*mu/3.)

    elif bulk and youngs:
        return 3. * bulk * (3*bulk - youngs) / (9*bulk - youngs)

    elif pr and mu:
        return 2. * pr * mu / (1 - 2.*pr)

    elif pr and youngs:
        return pr * youngs / ((1+pr) * (1-2*pr))

    elif pr and bulk:
        return 3. * bulk * pr / (1+pr)

    else:
        return None


def pmod(vp=None, vs=None, rho=None, pr=None, mu=None, lam=None, youngs=None, bulk=None):
    '''
    Computes P-wave modulus given either Vp, Vs, and rho, or
    any two elastic moduli (e.g. lambda and mu, or Young's
    and bulk moduli).

    SI units only.

    Args:
        vp, vs, and rho
        or any 2 from lam, mu, youngs, pr, and bulk

    Returns:
        P-wave modulus in pascals, Pa

    '''

    if vp and rho:
        return rho * vp**2

    elif lam and mu:
        return lam + 2*mu

    elif youngs and mu:
        return mu * (4.*mu - youngs) / (3.*mu - youngs)

    elif bulk and lam:
        return 3*bulk - 2.*lam

    elif bulk and mu:
        return bulk + (4.*mu/3.)

    elif bulk and youngs:
        return 3. * bulk * (3*bulk + youngs) / (9*bulk - youngs)

    elif lam and pr:
        return lam * (1 - pr) / pr

    elif pr and mu:
        return 2. * pr * mu * (1-pr) / (1 - 2.*pr)

    elif pr and youngs:
        return (1-pr) * youngs / ((1+pr) * (1 - 2.*pr))

    elif pr and bulk:
        return 3. * bulk * (1-pr) / (1+pr)

    else:
        return None


def vp(youngs=None, vs=None, rho=None, mu=None, lam=None, bulk=None, pr=None, pmod=None):
    '''
    Computes Vp given bulk density and any two elastic moduli
    (e.g. lambda and mu, or Young's and P moduli).

    SI units only.

    Args:
        Any 2 from lam, mu, youngs, pr, pmod, bulk
        Rho

    Returns:
        Vp in m/s

    '''

    if mu and lam and rho:
        return np.sqrt( (lam + 2.*mu) / rho )

    elif youngs and mu and rho:
        return np.sqrt( mu * (youngs - 4.*mu) / (rho * (youngs - 3.*mu)) )

    elif youngs and pr and rho:
        return np.sqrt( youngs * (1 - pr) / (rho * (1+pr)* (1 - 2.*pr)) )

    elif bulk and lam and rho:
        return np.sqrt( (9.*bulk - 2.*lam) / rho )

    elif bulk and mu and rho:
        return np.sqrt( (bulk + 4.*mu/3. ) / rho)

    elif lam and pr and rho:
        return np.sqrt( lam * (1. - pr) / (pr*rho))

    else:
        return None


def vs(youngs=None, vp=None, rho=None, mu=None, lam=None, bulk=None, pr=None, pmod=None):
    '''
    Computes Vs given bulk density and shear modulus.

    SI units only.

    Args:
        Mu
        Rho

    Returns:
        Vs in m/s

    '''

    if mu and rho:
        return np.sqrt( mu / rho )

    else:
        return None


def moduli_dict(vp, vs, rho):
    '''
    Computes elastic moduli given Vp, Vs, and rho.

    SI units only.

    Args:
        Vp, Vs, and rho

    Returns:
        A dict of elastic moduli, plus P-wave impedance.

    '''
    
    mod = {}

    mod['imp'] = vp * rho

    mod['mu'] = mu(vs=vs, rho=rho)
    mod['pr'] = pr(vp=vp, vs=vs, rho=rho)
    mod['lam'] = lam(vp=vp, vs=vs, rho=rho)
    mod['bulk'] = bulk(vp=vp, vs=vs, rho=rho)
    mod['pmod'] = pmod(vp=vp, rho=rho)
    mod['youngs'] = youngs(vp=vp, vs=vs, rho=rho)

    return mod



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
Seismic wavelets.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from scipy.signal import hilbert
from scipy.signal import chirp


def ricker(duration, dt, f):
    """
    Also known as the mexican hat wavelet, models the function:
    A =  (1-2 \pi^2 f^2 t^2) e^{-\pi^2 f^2 t^2}

    :param duration: The length in seconds of the wavelet.
    :param dt: is the sample interval in seconds (usually 0.001,
               0.002, 0.004)
    :params f: Center frequency of the wavelet (in Hz). If a list or tuple is
               passed, the first element will be used.

    :returns: ricker wavelets with center frequency f sampled at t.
    """


    freq = np.array(f)
     
    t = np.arange(-duration/2, duration/2 , dt)

    output = np.zeros((t.size, freq.size))
        
    for i in range(freq.size):
        pi2 = (np.pi ** 2.0)
        if ( freq.size == 1 ):
            fsqr = freq ** 2.0
        else:
            fsqr =  freq[i] ** 2.0
        tsqr = t ** 2.0
        pft = pi2 * fsqr * tsqr
        A = (1 - (2 * pft)) * np.exp(-pft)
        output[:,i] = A

    if freq.size == 1: output = output.flatten()
        
    return output / np.amax(output)
    
    
def sweep(duration, dt, f, method = 'linear', phi = 0,
          vertex_zero = True, autocorrelate = True):
    """
    Generates a linear frequency modulated wavelet (sweep)
    Does a wrapping of scipy.signal.chirp
    
    :param duration: The length in seconds of the wavelet.
    :param dt: is the sample interval in seconds (usually 0.001, 0.002, 0.004)
    :param f: Tuple of (f1, f2), or a similar list. A list of lists
              will create a wavelet bank.
    :keyword method: {'linear','quadratic','logarithmic'}, optional
    :keyword phi: float, phase offset in degrees
    :keyword vertex_zero: bool, optional
        This parameter is only used when method is 'quadratic'. 
        It determines whether the vertex of the parabola that 
        is the graph of the frequency is at t=0 or t=t1.

    :returns: An LFM waveform.
    """
    
    t = np.arange(-duration/2, duration/2 , dt) 
    t0 = -duration/2
    t1 = duration/2

    freq = np.array( f )

    if freq.size == 2:
         A = chirp(t, freq[0], t1, freq[1],
                    method, phi, vertex_zero)
         if autocorrelate:
             A = np.correlate(A, A, mode='same')
         output =  A / np.amax(A)
    
    else:
        output = np.zeros((t.size, freq.shape[1]))
        
        for i in range( freq.shape[1] ):
            A = chirp(t,freq[0,i],t1,freq[1,i],
                       method, phi, vertex_zero)
    
            if autocorrelate:
                A = np.correlate(A, A, mode='same')
            output[:,i] = A / np.max(A)
     
    return output


def ormsby(duration, dt, f):
    """
    The Ormsby wavelet requires four frequencies:
    f1 = low-cut frequency
    f2 = low-pass frequency
    f3 = high-pass frequency
    f4 = hi-cut frequency
    Together, the frequencies define a trapezoid shape in the
    spectrum.
    The Ormsby wavelet has several sidelobes, unlike Ricker wavelets
    which only have two, one either side.

    :param duration: The length in seconds of the wavelet.
    :param dt: is the sample interval in seconds (usually 0.001,
               0.002, 0.004)
    :params f: Tuple of form (f1,f2,f3,f4), or a similar list.

    :returns: A vector containing the ormsby wavelet
    """
    
    # Try to handle some duck typing
    if not (isinstance(f, list) or isinstance(f, tuple)):
        f = [f]
    
    # Deal with having fewer than 4 frequencies
    if len(f) == 4:
        f1 = f[0]
        f2 = f[1]
        f3 = f[2]
        f4 = f[3]
    else:
        # Cope with only having one frequency
        # This is an arbitrary hack, is this desirable?
        # Need a way to notify with warnings
        f1 = f[0]/4
        f2 = f[0]/2
        f3 = f[0]*2
        f4 = f[0]*2.5
    
    def numerator(f,t):
        return (np.sinc( f * t)**2) * ((np.pi * f) ** 2)

    pf43 = ( np.pi * f4 ) - ( np.pi * f3 )
    pf21 = ( np.pi * f2 ) - ( np.pi * f1 )
    
    t = np.arange(-duration/2, duration/2 , dt) 
    
    A = ((numerator(f4,t)/pf43) - (numerator(f3,t)/pf43) -
         (numerator(f2,t)/pf21) + (numerator(f1,t)/pf21))

    A /= np.amax(A)
    return A


def rotate_phase(w, phi):
    """
    Performs a phase rotation of wavelet using:
    A = w(t)Cos(phi) + h(t)Sin(phi)
    Where w(t) is the wavelet and h(t) is it's hilbert transform.
    
    :params w: The wavelet vector.
    :params phi: The phase rotation angle (in Radians) to apply.

    :returns: The phase rotated signal.
    """

    # Get the analytic signal for the wavelet
    a = hilbert(w, axis=0)

    A = (np.real(a) * np.cos(phi) +
         np.imag(a) * np.sin(phi))

    return A
    



from numpy import zeros, log2, ceil, arange, absolute, floor, sum
from scipy.fftpack import fft
from scipy.signal import get_window
from bruges.util import next_pow2
from numpy import hanning, concatenate

def spectra(data, window_length, dt=1.0, window_type='boxcar',
            overlap=0.5, normalize=False,
            zero_padding=0):
    """
    Calculates a spectrogram using windowed STFTs.
     
    :param data: 1D numpy array to process into spectra.
    :param window_length: The length of the window in seconds if
                          dt is set, otherwise in samples. Will
                          zero pad to the closest power of two.
    :keyword window_type: A string specifying the type of window to
                          use for the STFT. The same input as
                          scipy.signal.get_window
    :keyword dt: The time sample interval of the trace. Defaults to
                 1, which allows window_length to be in seconds.
                   
    :keyword overlap: The fractional overlap for each window.
                      A value of 0 uses no redudant data, a value of 1
                      slides the STFT window one sample at a time.
                      Defaults to 0.5
    :keyword normalize: Normalizes the each spectral slice to have
                        unity energy.
    :keyword zero_padding: The amount of zero padding to when creating
                           the spectra.
                      
    :returns: A spectrogram of the data ([time, freq]).
            ( 2D array for 1D input )
    """

    
    # Make the base window
    window_n = floor( window_length / dt )
    pad = floor( zero_padding / dt )
    window = get_window(window_type, window_n)
    
    # Calculate how many STFTs we need to do.
    stride = window.size * (1 - overlap) + 1
    n_windows = int(ceil((data.size - window.size) / stride) + 1)

    # Pad window to power of 2
    padded_window = zeros(next_pow2(window.size+pad))
    
    # Init the output array
    output = zeros([n_windows, int(padded_window.size)/2])

    norm_factor = 1.0
    
    # Do the loop
    for i in range(int(n_windows)):

        start = int(i * stride)
        end = start + int(window.size)
        
        # Do not compute last samples if we don't have a full window
        if ( end > data.size-1 ): break

        padded_window[0:window.size] = window*data[start:end]
        spect = (2.* absolute(fft(padded_window))/ \
                         window.size)[0:int(padded_window.size /2.)]
                         
        if normalize: output[i, :] = spect / sum(spect**2)
                
        else: output[i, :] = spect
    
    return output



    
        
        
        
        



import numpy as np
from bruges.attribute import energy 

def dipsteer( data, window_length, stepout,
              maxlag, overlap=0., dt=1 ):
    """
    Calculates a dip field by finding the maximum correlation between
    adjacent traces.

    :param data: A 2D seismic section (samples,traces) used to
                 calculate dip.
    :param window_length: The length [seconds] of the window to use
    :param dt: The time sample interval of the traces.
                          
    :param stepout: The number of traces on either side of each point
                    to average when calculating the dip.
    :param maxlag: The maximum amount time lag to use when correlating
                   the traces.
    :keyword overlap: The fractional overlap for each window.
                      A value of 0 uses no redudant data, a value of 1
                      slides the dip correlator one sample at a time.
                      Defaults to 0.5
    

    :returns: a dip field [samples/trace] of the same shape as the
             input data, and correlation coefficients corresponding
             to the data.
    """

    dip = np.zeros( data.shape )

    window_length = np.floor( window_length / dt )
    
    # Force the window length to be odd for index tracking
    if not ( window_length % 2 ): window_length+=1
    
    # Define time windows
    stride = window_length * ( 1 - overlap ) 
    n_windows = np.ceil( ( data.shape[0] - window_length ) /
                         stride ) + 1
 
    # Normalize each trace to the same RMS energy
    norm_factor = np.sqrt( energy( data, window_length ) )
    norm_data = data / norm_factor

    # Replace the 0/0 with 0
    norm_data = np.nan_to_num( norm_data )

    # Mid point in the data which corresponds to zero dip
    zero_dip = ( np.floor( window_length / 2.0) +  maxlag  )
    
    # Loop over each trace we can do a full calculation for
    for i in np.arange( stepout +1, data.shape[-1] - (stepout +1) ):
        
        # Loop over each time window
        for j in np.arange( 0, n_windows):

            start = (j * stride) + ( maxlag )
            end = start + window_length

            # Do not compute last samples if we don't
            # have a full window
            if ( end > (norm_data.shape[0]-maxlag) ): break
        
            # Get the  kernel
            kernel = norm_data[ start : end, i ]

            dips_j = 0
         
            # Correlate with adjacent traces
            for k in np.arange( 1, stepout + 1 ):

                # Do the trace on the right
                r_trace = norm_data[ start - (k*maxlag) : \
                                     end + (k*maxlag), i+k ]
               
                cor_r = np.correlate( kernel, r_trace, mode='same' )

                if ( np.amax( cor_r ) < .1 ): dip_r = 0
                else:
                    dip_r = (  np.argmax( cor_r ) - zero_dip ) / k

               
                # Do the left trace
                l_trace = norm_data[ start - (k*maxlag) : \
                                     end + (k*maxlag), i-k ]
                
                cor_l = np.correlate( kernel, l_trace, mode = 'same' )

                if ( np.amax( cor_l ) < .1 ): dip_l = 0
                else:
                    dip_l = -(  np.argmax( cor_l ) - zero_dip ) / k

          
                dips_j += dip_r
                dips_j += dip_l

              
            # Average the result
            dips_j /= ( 2. * stepout )

            # Update the output 
            dip[ start : start + stride, i ] = dips_j
        
    return dip





from scipy.interpolate import interp1d
from numpy import arange, amax, amin, asarray, zeros, cumsum, \
     transpose, gradient, mean, size


def __convert(data, vmodel, interval, interval_new, scale, mode):
    """
    Generic function for converting between scales. Use either
    time to depth or depth to time
    """

    dz = interval
    dt = interval_new

    if len(data.shape) == 1:
        ntraces = 1
        nsamps = data.size
        if size(dz) == 1:
            depths = arange(nsamps)*dz
        else:
            depths = dz
        v_avg = cumsum(vmodel) / (arange(nsamps) + 1)

    else:
        ntraces = data.shape[-1]
        nsamps = data.shape[0]
        if size(dz) == 1:
            tr = [(arange(nsamps) * dz) for i in range(ntraces)]
            depths = transpose(asarray(tr))
        else:
            depths = dz
        tr = [arange(nsamps) + 1 for i in range(ntraces)]
        v_avg = cumsum(vmodel, axis=0) / transpose(tr)

    times = depths / v_avg
    times *= scale

    if size(dt) == 1:
        times_lin = arange(amin(times), amax(times), dt)
    else:
        times_lin = dt

    if ntraces == 1:
        inter = interp1d(times, data,
                         bounds_error=False,
                         fill_value=data[-1],
                         kind='nearest')
        return inter(times_lin)
    else:
        output = zeros((times_lin.size, ntraces))
        for i in range(ntraces):
            inter = interp1d(times[:, i], data[:, i],
                             bounds_error=False,
                             fill_value=data[-1, i],
                             kind=mode)
            output[:, i] += inter(times_lin)
        return output


def time_to_depth(data, vmodel, dt, dz, twt=True,
                  mode="nearest"):
    """
    Converts data from the time domain to the depth domain given a
    velocity model.

    :param data: The data to convert, will work with a 1 or 2D numpy
                 numpy array. array(samples,traces).
    :param vmodel: P-wave velocity model that corresponds to the data.
                   Must be the same shape as data.
    :param dt: The sample interval of the input data [s], or an
               array of times.
    :param dz: The sample interval of the output data [m], or an
               array of depths

    :keyword twt: Use twt travel time, defaults to true

    :returns: The data resampled in the depth domain.
    """

    if twt:
        scale = 1/2.0
    else:
        scale = 1.0
    # Do depth to time with inverted velocity profile
    return __convert(data, 1. / vmodel, dt, dz, scale, mode)


def depth_to_time(data, vmodel, dz, dt, twt=True, mode="nearest"):
    """
    Converts data from the depth domain to the time domain given a
    velocity model.

    :param data: The data to convert, will work with a 1 or 2D numpy
                 numpy array. array(samples,traces).
    :param vmodel: P-wave velocity model that corresponds to the data.
                   Must be the same shape as data.
    :param dz: The sample interval of the input data [m].
    :param dt: The sample interval of the output data [s].

    :keyword twt: Use twt travel time, defaults to true

    :returns: The data resampled in the time domain.
    """

    if twt:
        scale = 2.0
    else:
        scale = 1.0
    # Do depth to time with inverted velocity profile
    return __convert(data, vmodel, dz, dt, scale, mode)
